/*                                                                              *
 *   This file is part of the ESO VISIR package                                 *
 *   Copyright (C) 2004,2005 European Southern Observatory                      *
 *                                                                              *
 *   This library is free software; you can redistribute it and/or modify       *
 *   it under the terms of the GNU General Public License as published by       *
 *   the Free Software Foundation; either version 2 of the License, or          *
 *   (at your option) any later version.                                        *
 *                                                                              *
 *   This program is distributed in the hope that it will be useful,            *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 *   GNU General Public License for more details.                               *
 *                                                                              *
 *   You should have received a copy of the GNU General Public License          *
 *   along with this program; if not, write to the Free Software                *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA       *
 *                                                                              */
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <visir_inputs.h>

/*----------------------------------------------------------------------------*/
/**
 * @defgroup visir_inputs_test Testing of the VISIR utilities
 */
/*----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
   @brief   Unit tests of inputs module
**/
/*----------------------------------------------------------------------------*/

static void visir_img_check_box_test(void);
static void visir_img_check_align_test(void);

#ifndef BOX_SIZE
#define BOX_SIZE 11
#endif

#ifndef LINE_SIZE
#define LINE_SIZE BOX_SIZE
#endif

int main(void)
{

    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);

    visir_img_check_align_test();
    visir_img_check_box_test();

    return cpl_test_end(0);

}

static void visir_img_check_box_test(void)
{

    cpl_image     * image;
    const double    sigma = 1.0;
    cpl_apertures * appos;
    cpl_apertures * apneg;
    double          ecc;
    cpl_boolean     swapp, swapn;
    FILE          * stream;
    cpl_error_code  error = CPL_ERROR_NONE;
    int i, j;

    stream = cpl_msg_get_level() > CPL_MSG_INFO
        ? fopen("/dev/null", "a") : stdout;

    cpl_test_nonnull( stream );

    image = cpl_image_new(3 * BOX_SIZE, 3 * BOX_SIZE, CPL_TYPE_INT);

    /* Create four 3x3-pixel objects */
    for (j = -1; j <= 1; j++) {
      for (i = -1; i <= 1; i++) {
          error |= cpl_image_set(image,   BOX_SIZE + i,   BOX_SIZE + j, -1.0);
          error |= cpl_image_set(image, 2*BOX_SIZE + i, 2*BOX_SIZE + j, -1.0);
          error |= cpl_image_set(image,   BOX_SIZE + i, 2*BOX_SIZE + j,  1.0);
          error |= cpl_image_set(image, 2*BOX_SIZE + i,   BOX_SIZE + j,  1.0);
          cpl_test_eq_error(error, CPL_ERROR_NONE);
      }
    }

    appos = cpl_apertures_extract_sigma(image, sigma);
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_nonnull(appos);
    cpl_test_eq(cpl_apertures_get_size(appos), 2);

    cpl_apertures_dump(appos, stream);

    error = cpl_image_multiply_scalar(image, -1.0);
    cpl_test_eq_error(error, CPL_ERROR_NONE);

    apneg = cpl_apertures_extract_sigma(image, sigma);
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_nonnull(apneg);
    cpl_test_eq(cpl_apertures_get_size(apneg), 2);

    cpl_apertures_dump(apneg, stream);

    /* Various error checks */
    ecc = visir_img_check_box(NULL, 1, 2, apneg, 1, 2, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_NULL_INPUT);

    ecc = visir_img_check_box(appos, 1, 2, NULL, 1, 2, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_NULL_INPUT);

    ecc = visir_img_check_box(appos, 1, 2, apneg, 1, 2, BOX_SIZE,
                              &swapp, NULL);
    cpl_test_error(CPL_ERROR_UNSPECIFIED);

    ecc = visir_img_check_box(appos, 1, 2, apneg, 1, 2, BOX_SIZE,
                              NULL, &swapn);
    cpl_test_error(CPL_ERROR_UNSPECIFIED);

    ecc = visir_img_check_box(appos, 1, 2, appos, 1, 2, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_UNSPECIFIED);

    ecc = visir_img_check_box(appos, 0, 2, apneg, 1, 2, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

    ecc = visir_img_check_box(appos, 1, 2, apneg, 1, 0, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

    ecc = visir_img_check_box(appos, 1, 2, apneg, 1, 3, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

    ecc = visir_img_check_box(appos, 3, 2, apneg, 1, 2, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

    ecc = visir_img_check_box(appos, 1, 2, apneg, 1, 2, 0.0, &swapp, &swapn);
    cpl_test_error(CPL_ERROR_UNSPECIFIED);

    /* Verify output of correct call - with zero eccentricity */
    ecc = visir_img_check_box(appos, 1, 2, apneg, 1, 2, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_abs(ecc, 0.0, FLT_EPSILON);
    cpl_test(swapp);
    cpl_test_zero(swapn);

    /* - and points are swapped when needed */
    ecc = visir_img_check_box(appos, 2, 1, apneg, 2, 1, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_abs(ecc, 0.0, FLT_EPSILON);
    cpl_test_zero(swapp);
    cpl_test(swapn);

    /* Verify output of correct call - with maximum (?) eccentricity */
    ecc = visir_img_check_box(apneg, 1, 2, appos, 1, 2, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_abs(ecc, 2.0 * CPL_MATH_SQRT2, FLT_EPSILON);
    cpl_test_zero(swapp);
    cpl_test(swapn);

    /* - and points are swapped when needed */
    ecc = visir_img_check_box(apneg, 2, 1, appos, 2, 1, BOX_SIZE,
                              &swapp, &swapn);
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_abs(ecc, 2.0 * CPL_MATH_SQRT2, FLT_EPSILON);
    cpl_test(swapp);
    cpl_test_zero(swapn);

    cpl_image_delete(image);
    cpl_apertures_delete(appos);
    cpl_apertures_delete(apneg);

    if (stream != stdout) cpl_test_zero( fclose(stream) );

}

static void visir_img_check_align_test(void)
{

    FILE          * stream;
    cpl_error_code  error = CPL_ERROR_NONE;
    int             idir;
    cpl_boolean     is_hor = CPL_FALSE;

    stream = cpl_msg_get_level() > CPL_MSG_INFO
        ? fopen("/dev/null", "a") : stdout;

    cpl_test_nonnull( stream );

    for (idir = 0; idir < 2; idir++, is_hor = CPL_TRUE) {
        cpl_image     * image = cpl_image_new(4 * LINE_SIZE, 4 * LINE_SIZE,
                                              CPL_TYPE_INT);
        const double    sigma = 1.0;
        cpl_apertures * appos;
        cpl_apertures * apneg;
        double          ecc;
        cpl_boolean     swapn, swapn2;
        int i, j;


        /* Create three 3x3-pixel objects */
        for (j = -1; j <= 1; j++) {
            for (i = -1; i <= 1; i++) {
                error |= cpl_image_set(image, 2*LINE_SIZE + i,
                                       2*LINE_SIZE + j,  1.0);

                if (is_hor) {
                    error |= cpl_image_set(image,   LINE_SIZE + i,
                                           2*LINE_SIZE + j, -1.0);
                    error |= cpl_image_set(image, 3*LINE_SIZE + i,
                                           2*LINE_SIZE + j, -1.0);
                } else {
                    error |= cpl_image_set(image, 2*LINE_SIZE + i,
                                           LINE_SIZE + j, -1.0);
                    error |= cpl_image_set(image, 2*LINE_SIZE + i,
                                           3*LINE_SIZE + j, -1.0);
                }
                cpl_test_eq_error(error, CPL_ERROR_NONE);
            }
        }

        appos = cpl_apertures_extract_sigma(image, sigma);
        cpl_test_error(CPL_ERROR_NONE);
        cpl_test_nonnull(appos);
        cpl_test_eq(cpl_apertures_get_size(appos), 1);

        cpl_apertures_dump(appos, stream);

        error = cpl_image_multiply_scalar(image, -1.0);
        cpl_test_eq_error(error, CPL_ERROR_NONE);

        apneg = cpl_apertures_extract_sigma(image, sigma);
        cpl_test_error(CPL_ERROR_NONE);
        cpl_test_nonnull(apneg);
        cpl_test_eq(cpl_apertures_get_size(apneg), 2);

        cpl_apertures_dump(apneg, stream);

        /* Various error checks */
        ecc = visir_img_check_align(NULL, 1, apneg, 1, 2, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_NULL_INPUT);

        ecc = visir_img_check_align(appos, 1, NULL, 1, 2, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_NULL_INPUT);

        ecc = visir_img_check_align(appos, 1, apneg, 1, 2, LINE_SIZE, is_hor,
                                    NULL);
        cpl_test_error(CPL_ERROR_UNSPECIFIED);

        ecc = visir_img_check_align(appos, 0, apneg, 1, 2, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

        ecc = visir_img_check_align(appos, 1, apneg, 1, 0, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

        ecc = visir_img_check_align(appos, 1, apneg, 0, 1, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

        ecc = visir_img_check_align(appos, 1, apneg, 1, 3, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

        ecc = visir_img_check_align(appos, 3, apneg, 1, 2, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);

        ecc = visir_img_check_align(appos, 1, apneg, 1, 2, 0.0, is_hor, &swapn);
        cpl_test_error(CPL_ERROR_UNSPECIFIED);

        /* Verify output of correct call - with zero eccentricity */
        ecc = visir_img_check_align(appos, 1, apneg, 1, 2, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_NONE);

        cpl_test_abs(ecc, 0.0, FLT_EPSILON);
        cpl_test_zero(swapn);

        /* - and points are swapped when needed */
        ecc = visir_img_check_align(appos, 1, apneg, 2, 1, LINE_SIZE, is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_NONE);

        cpl_test_abs(ecc, 0.0, FLT_EPSILON);
        cpl_test(swapn);

        /* Verify output of correct call - with maximum (?) eccentricity */
        ecc = visir_img_check_align(appos, 1, apneg, 1, 2, LINE_SIZE, !is_hor,
                                    &swapn);
        cpl_test_error(CPL_ERROR_NONE);

        cpl_test_abs(ecc, 2.0, FLT_EPSILON);

        /* - also with points swapped */
        ecc = visir_img_check_align(appos, 1, apneg, 2, 1, LINE_SIZE, !is_hor,
                                    &swapn2);
        cpl_test_error(CPL_ERROR_NONE);

        cpl_test_abs(ecc, 2.0, FLT_EPSILON);

        /* FIXME: When to exactly swap here is not well-defined. */

        cpl_image_delete(image);
        cpl_apertures_delete(appos);
        cpl_apertures_delete(apneg);
    }

    if (stream != stdout) cpl_test_zero( fclose(stream) );

}
